(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2013 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uindex_set_types.ml,v 1.1 2013/06/04 17:11:41 fclement Exp $ *)

(* The private type of set of multi-indices. *)

type 'a multi_d = 'a list;;

(* Mono-dimensional range. *)

type mono_range =
  {
    left : Unatural.t;
    right : Unatural.t;
  }
;;

let empty_mono_range = { left = Unatural._0_; right = Unatural._0_; };;

let is_empty_mono_range r = r = empty_mono_range;;

let mono_range l r =
  if r < l then
    invalid_arg
      "Uindex_set.mono_range: second argument is lesser than the first" else
  if r = l then empty_mono_range else
  { left = l; right = r; }
;;

(* Multi-dimensional range. *)

type multi_range = mono_range multi_d;;

let empty_multi_range = [];;

let is_empty_multi_range r = r = empty_multi_range;;

let multi_range mrs =
  if Ulist.exists is_empty_mono_range mrs then empty_multi_range else mrs
;;

(* Index set. *)

type index = int multi_d;;

type index_set =
  | Range of multi_range
  | Extensional of index array
  | Union of index_set * index_set
;;

let empty_index_set = Range empty_multi_range;;

let is_empty_index_set is = is = empty_index_set;;

let rec reference_point = function
  | Range mr -> Ulist.map (fun { left = l; _; } -> l) mr
  | Extensional ia -> ia.(0)
  | Union (is1, _) -> reference_point is1
;;

let compare is1 is2 =
  Pervasives.compare (reference_point is1) (reference_point is2)
;;

let range mr = Range mr;;

let extensional ia =
  let len = Array.length ia in
  if len = 0 then empty_index_set else
  let ia = Array.sort ia in
  let x = ia.(0) in
  let rec loop before_mr l r after_mr i = ... in
  loop [] x (succ x) [] 1
;;

let rec range_range_union ri1 ri2 =
  let ({ left = l1; right = r1; },
       { left = l2; right = r2; }) = (ri1, ri2) in
  if r1 < l2 then Union (ri1, ri2) else
  range l1 (Unatural.max r1 r2)
;;

let rec range_range_union mr1 mr2 =
  let rec loop

;;

let range_extensional_union mr1 ia2 =
  let { left = l1; right = r1; } = mr1 in
  let len2 = Array.length ia2 in
  let l2 = ia2.(0)
  and r2 = ia2.(len2 - 1) + 1 in
  if r1 < l2 then Union (range mr1, extensional ia2) else
  if r2 < l1 then Union (extensional ia2, range mr1) else
  let rec loop before_mr1 l1 r1 after_mr1 i =
    if i >= len2
    then
      let mr1 = range (range l1 r1) in
      match (before_mr1, after_mr1) with
      | ([], []) -> mr1
      | ([], after_mr1) ->
          Union (mr1, Extensional (Array.of_list (List.rev after_mr1)))
      | (before_mr1, []) ->
          Union (Extensional (Array.of_list (List.rev before_mr1)), mr1)
      | (before_mr1, after_mr1) ->
          Union
            (Extensional (Array.of_list (List.rev before_mr1)),
             Union (mr1, Extensional (Array.of_list (List.rev after_mr1))))
    else
      let x = ia2.(i) in
      if x < Unatural.pred l1 then
        loop (x :: before_mr1) l1 r1 after_mr1 (i + 1) else
      if x = Unatural.pred l1 then
        loop before_mr1 x r1 after_mr1 (i + 1) else
      if x < Unatural.succ r1 then
        loop before_mr1 l1 r1 after_mr1 (i + 1) else
      if x = Unatural.succ r1 then
        loop before_mr1 l1 x after_mr1 (i + 1) else
      loop before_mr1 l1 r1 (x :: after_mr1) (i + 1) in
  loop [] l1 r1 [] 0
;;

let extensional_extensional_union ia1 ia2 =
  extensional (Array.concat ia1 ia2)
;;

let rec union is1 is2 =
  if is_empty_index_set is1 || is_empty_index_set is2 then empty_index_set else
  if compare is1 is2 > 0 then union is2 is1 else
  match (is1, is2) with
  | (Range mr1, Range mr2) -> range_range_union mr1 mr2
  | (Range mr1, Extensional ia2) -> range_extensional_union mr1 ia2
  | (Extensional ia1, Range mr2) -> range_extensional_union mr2 ia1
  | (Extensional ia1, Extensional ia2) -> extensional_extensional_union ia1 ia2
  | (Range _, Union (is21, is22))
  | (Extensional _, Union (is21, is22)) ->
      union (union is1 is21) is22
  | (Union (is11, is12), Range _)
  | (Union (is11, is12), Extensional _)
  | (Union (is11, is12), Union (_, _)) ->
      union is11 (union is12 is2)
;;
